/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:          ctrec.cpp
**   Purpose:       Reconstruct an image from projections
**   Programmer:    Kevin Rosenberg
**   Date Started:  Aug 1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: ctrec.cpp,v 1.4 2000/06/08 16:43:10 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"

enum {O_INTERP, O_FILTER, O_FILTER_PARAM, O_BACKPROJ, O_VERBOSE, O_TRACE, O_HELP, O_DEBUG, O_VERSION};

static struct option my_options[] =
{
  {"interp", 1, 0, O_INTERP},
  {"filter", 1, 0, O_FILTER},
  {"filter-param", 1, 0, O_FILTER_PARAM},
  {"backproj", 1, 0, O_BACKPROJ},
  {"trace", 1, 0, O_TRACE},
  {"debug", 0, 0, O_DEBUG},
  {"verbose", 0, 0, O_VERBOSE},
  {"help", 0, 0, O_HELP},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
ctrec_usage (const char *program)
{
  fprintf(stdout,"usage: %s raysum-file image-file nx-image ny-image [OPTIONS]\n", kbasename(program));
  fprintf(stdout,"Image reconstruction from raysum projections\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"   raysum-file     Input raysum file\n");
  fprintf(stdout,"   image-file      Output image file in SDF2D format\n");
  fprintf(stdout,"   nx-image        Number of columns in output image\n");
  fprintf(stdout,"   ny-image        Number of rows in output image\n");
  fprintf(stdout,"   --interp        Interpolation method during backprojection\n");
  fprintf(stdout,"       nearest     Nearest neighbor interpolation\n");
  fprintf(stdout,"       linear      Linear interpolation\n");
#if HAVE_BSPLINE_INTERP
  fprintf(stdout,"       bspline     B-spline interpolation\n");
#endif
  fprintf(stdout,"    --filter       Filter name\n");
  fprintf(stdout,"       abs_bandlimit Abs * Bandlimiting (default)\n");
  fprintf(stdout,"       abs_sinc      Abs * Sinc\n");
  fprintf(stdout,"       abs_cos       Abs * Cosine\n");
  fprintf(stdout,"       abs_hamming   Abs * Hamming\n");
  fprintf(stdout,"       shepp         Shepp-Logan\n");
  fprintf(stdout,"       bandlimit     Bandlimiting\n");
  fprintf(stdout,"       sinc          Sinc\n");
  fprintf(stdout,"       cos           Cosine\n");
  fprintf(stdout,"       triangle      Triangle\n");
  fprintf(stdout,"       hamming       Hamming\n");
  fprintf(stdout,"    --backproj     Backprojection Method\n");
  fprintf(stdout,"       trig        Trigometric functions at every point\n");
  fprintf(stdout,"       table       Trigometric functions with precalculated table\n");
  fprintf(stdout,"       diff        Difference method\n");
  fprintf(stdout,"       diff2       Optimized difference method (default)\n");
  fprintf(stdout,"       idiff2      Optimized difference method with integer math\n");
  fprintf(stdout,"    --filter-param Alpha level for Hamming filter\n");
  fprintf(stdout,"    --trace        Set tracing to level\n");
  fprintf(stdout,"         none      No tracing (default)\n");
  fprintf(stdout,"         text      Text level tracing\n");
  fprintf(stdout,"         phm       Trace phantom\n");
  fprintf(stdout,"         rays      Trace allrays\n");
  fprintf(stdout,"         plot      Trace plotting\n");
  fprintf(stdout,"         clipping  Trace clipping\n");
  fprintf(stdout,"    --verbose      Turn on verbose mode\n");
  fprintf(stdout,"    --debug        Turn on debug mode\n");
  fprintf(stdout,"    --version      Print version\n");
  fprintf(stdout,"    --help         Print this help message\n");
}


#ifdef HAVE_MPI
static void mpi_scatter_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int debug);
#endif

static void print_raysum_info(const RAYSUM *rs);

int 
ctrec_main (const int argc, char *const argv[])
{
  ImageFile *im_global = NULL;
  RAYSUM *rs_global = NULL;
  char *rs_name, *im_filename = NULL;
  char remark[MAXREMARK];
  char filt_name[80];
  double time_start = 0, time_end = 0;
  char *endptr;
  int opt_verbose = 0;
  int opt_debug = 0;
  int opt_trace = TRACE_NONE;
  double opt_filter_param = 1;
  FilterType opt_filter = FILTER_ABS_BANDLIMIT;
  InterpolationType opt_interp = I_LINEAR;
  int opt_interp_param = 1;
  BackprojType opt_backproj = O_BPROJ_DIFF2;
  int nx, ny;
#ifdef HAVE_MPI
  ImageFile *im_local;
  RAYSUM *rs_local;
  int mpi_argc = argc;
  char **mpi_argv = (char **) argv;
  int mpi_nview, mpi_ndet;
  double mpi_detinc, mpi_rotinc, mpi_phmlen;
  double mpi_t1, mpi_t2, mpi_t, mpi_t_g;

  MPI::Init (mpi_argc, mpi_argv);
  mpi_ct.comm = MPI::COMM_WORLD.Dup();
  mpi_ct.nproc = mpi_ct.comm.Get_size();
  mpi_ct.my_rank = mpi_ct.comm.Get_rank();

  if (mpi_ct.nproc > CT_MPI_MAX_PROCESS) {
    sys_error(ERR_FATAL, "Number of mpi processes (%d) exceeds max processes (%d)",
	      mpi_ct.nproc, CT_MPI_MAX_PROCESS);
    exit(1);
  }
#endif

#ifdef HAVE_MPI
  time_start = MPI::Wtime();
#else
  time_start = td_current_sec();
#endif

#ifdef HAVE_MPI
  if (mpi_ct.my_rank == 0) {
#endif
    while (1) {
      int c = getopt_long(argc, argv, "", my_options, NULL);
      char *endptr = NULL;
      
      if (c == -1)
	break;
      
      switch (c)
	{
	case O_INTERP:
	  if ((opt_interp = opt_set_interpolation(optarg)) < 0) {
	    ctrec_usage(argv[0]);
	    return (1);
	  }
	  break;
	case O_FILTER:
	  if ((opt_filter = opt_set_filter(optarg)) < 0) {
	    ctrec_usage(argv[0]);
	    return (1);
	  }
	  break;
	case O_BACKPROJ:
	  if ((opt_backproj = opt_set_backproj(optarg)) < 0) {
	      ctrec_usage(argv[0]);
	      return (1);
	  }
	  break;
	case O_FILTER_PARAM:
	  opt_filter_param = strtod(optarg, &endptr);
	  if (endptr != optarg + strlen(optarg)) {
	    ctrec_usage(argv[0]);
	  }
	  break;
	case O_VERBOSE:
	  opt_verbose = 1;
	  break;
	case O_DEBUG:
	  opt_debug = 1;
	  break;
	case O_TRACE:
	  if ((opt_trace = opt_set_trace(optarg)) < 0) {
	    ctrec_usage(argv[0]);
	    return (1);
	  }
	  break;
        case O_VERSION:
#ifdef VERSION
	  fprintf(stdout, "Version %s\n", VERSION);
#else
	  fprintf(stderr, "Unknown version number");
#endif
	  exit(0);
	case O_HELP:
	case '?':
	  ctrec_usage(argv[0]);
	  return (0);
	default:
	  ctrec_usage(argv[0]);
	  return (1);
	}
    }
  
    if (optind + 4 != argc) {
      ctrec_usage(argv[0]);
      return (1);
    }

    rs_name = argv[optind];
  
    im_filename = argv[optind + 1];
  
    nx = strtol(argv[optind + 2], &endptr, 10);
    ny = strtol(argv[optind + 3], &endptr, 10);
  
    if (opt_filter == FILTER_G_HAMMING || opt_filter == FILTER_ABS_G_HAMMING)
      snprintf (filt_name, sizeof(filt_name), "%s: alpha = %.2f",
	       name_of_filter (opt_filter), opt_filter_param); 
    else
      snprintf (filt_name, sizeof(filt_name), "%s", name_of_filter (opt_filter));
  
    snprintf (remark, sizeof(remark), "Reconstruct: %dx%d, %s, %s, %s",
	     nx, ny, filt_name, name_of_interpolation (opt_interp), name_of_backproj(opt_backproj));
  
    if (opt_verbose)
      fprintf (stdout, "%s\n", remark);
#ifdef HAVE_MPI
  }
#endif

#ifdef HAVE_MPI
  if (mpi_ct.my_rank == 0) {
    rs_global = raysum_open (rs_name);
    raysum_read (rs_global);
    if (opt_verbose)
      print_raysum_info(rs_global);

    mpi_ndet = rs_global->ndet;
    mpi_nview = rs_global->nview;
    mpi_detinc = rs_global->det_inc;
    mpi_phmlen = rs_global->phmlen;
    mpi_rotinc = rs_global->rot_inc;
  }

  mpi_t1 = MPI::Wtime();
  mpi_ct.comm.Bcast (&opt_verbose, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_debug, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_trace, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_filter, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_interp, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_filter_param, 1, MPI::DOUBLE, 0);
  mpi_ct.comm.Bcast (&opt_interp_param, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&opt_backproj, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&mpi_ndet, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&mpi_nview, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&mpi_detinc, 1, MPI::DOUBLE, 0);
  mpi_ct.comm.Bcast (&mpi_phmlen, 1, MPI::DOUBLE, 0);
  mpi_ct.comm.Bcast (&mpi_rotinc, 1, MPI::DOUBLE, 0);
  mpi_ct.comm.Bcast (&nx, 1, MPI::INT, 0);
  mpi_ct.comm.Bcast (&ny, 1, MPI::INT, 0);
  if (opt_verbose) {
    mpi_t2 = MPI::Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    mpi_ct.comm.Reduce(&mpi_t, &mpi_t_g, 1, MPI::DOUBLE, MPI::MAX, 0);
    if (mpi_ct.my_rank == 0)
      printf("Time to Bcast vars = %f secs, Max time = %f\n", mpi_t, mpi_t_g);
  }

  mpi_ct_calc_work_units(mpi_nview);

  rs_local = raysum_create (NULL, mpi_ct.local_work_units[mpi_ct.my_rank], mpi_ndet);

  rs_local->ndet = mpi_ndet;
  rs_local->nview = mpi_nview;
  rs_local->det_inc = mpi_detinc;
  rs_local->phmlen = mpi_phmlen;
  rs_local->rot_inc = mpi_rotinc;

  if (opt_verbose)
    mpi_t1 = MPI::Wtime();
  mpi_scatter_rs(rs_global, rs_local, opt_debug);
  if (opt_verbose) {
    mpi_t2 = MPI::Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    mpi_ct.comm.Reduce(&mpi_t, &mpi_t_g, 1, MPI::DOUBLE, MPI::MAX, 0);
    if (mpi_ct.my_rank == 0)
      printf("Time to scatter rs = %f secs, Max time = %f sec\n", mpi_t, mpi_t_g);
  }

  if (mpi_ct.my_rank == 0) {
    im_global = new ImageFile (im_filename, nx, ny);
    im_global->adf.fileCreate();
  }

  im_local = new ImageFile (nx, ny);
#else
  rs_global = raysum_open (rs_name);
  raysum_read (rs_global);
  if (opt_verbose)
    print_raysum_info(rs_global);

  im_global = new ImageFile (im_filename, nx, ny);
  im_global->adf.fileCreate();
#endif

#ifdef HAVE_MPI
  mpi_t1 = MPI::Wtime();
  proj_reconst (*im_local, rs_local, opt_filter, opt_filter_param, 
		 opt_interp, opt_interp_param, opt_backproj, opt_trace);

  mpi_t2 = MPI::Wtime();
  mpi_t = mpi_t2 - mpi_t1;
  mpi_ct.comm.Reduce(&mpi_t, &mpi_t_g, 1, MPI::DOUBLE, MPI::MAX, 0);
  if (mpi_ct.my_rank == 0 && opt_verbose)
    printf("Time to reconstruct = %f, Max time = %f\n", mpi_t, mpi_t_g);
#else
  proj_reconst (*im_global, rs_global, opt_filter, opt_filter_param, 
		 opt_interp, opt_interp_param, opt_backproj, opt_trace);
#endif

#ifdef HAVE_MPI
  if (opt_verbose)
    mpi_t1 = MPI::Wtime();

  int nxLocal = im_local->adf.nx();
  int nyLocal = im_local->adf.ny();
  ImageFileArray vLocal = im_local->getArray();
  ImageFileArray vGlobal = NULL;
  if (mpi_ct.my_rank == 0)
    vGlobal = im_global->getArray();

  for (int ix = 0; ix < nxLocal; ix++) {
    void *recvbuf = NULL;
    if (mpi_ct.my_rank == 0)
      recvbuf = vGlobal[ix];

    mpi_ct.comm.Reduce(vLocal[ix], recvbuf, nyLocal, im_local->getMPIDataType(), MPI::SUM, 0);
  }

  if (opt_verbose) {
    mpi_t2 = MPI::Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    mpi_ct.comm.Reduce (&mpi_t, &mpi_t_g, 1, MPI::DOUBLE, MPI::MAX, 0);
    if (mpi_ct.my_rank == 0)
      printf("Time to reduce image = %f secs, max time = %f\n", mpi_t, mpi_t_g);
  }

  if (mpi_ct.my_rank == 0)
      time_end = MPI::Wtime();
#else
  time_end = td_current_sec();
#endif


#ifdef HAVE_MPI
  if (mpi_ct.my_rank == 0)
#endif
    {
      raysum_close (rs_global);
      double calctime = time_end - time_start;
      im_global->adf.arrayDataWrite ();
      im_global->adf.labelAdd (Array2dFileLabel::L_HISTORY, rs_global->remark, rs_global->calctime);
      im_global->adf.labelAdd (Array2dFileLabel::L_HISTORY, remark, calctime);
      im_global->adf.fileClose ();
      if (opt_verbose)
	cout << "Time active = " << calctime << " sec" << endl;
    }

#ifdef HAVE_MPI
	MPI::Finalize();
#endif

	return (0);
}


#ifdef HAVE_MPI
static void mpi_scatter_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int opt_debug)
{
  int iproc;
  int end_work_unit;
  int iw = 0;

  if (mpi_ct.my_rank == 0) {
    for (iproc = 0; iproc < mpi_ct.nproc; iproc++) {
      end_work_unit = mpi_ct.start_work_unit[iproc] + mpi_ct.local_work_units[iproc]  - 1;

      for (iw = mpi_ct.start_work_unit[iproc]; iw <= end_work_unit; iw++) {
	mpi_ct.comm.Send(&rs_global->view[iw]->ndet, 1, MPI::INT, iproc, 0);
	mpi_ct.comm.Send(&rs_global->view[iw]->view_angle, 1, MPI::DOUBLE, iproc, 0);
	mpi_ct.comm.Send(rs_global->view[iw]->detval, rs_global->ndet, MPI::FLOAT, iproc, 0);
      }
    }
  }

  if (opt_debug)
    fprintf(stdout, "Receiving rs data in process %d\n", mpi_ct.my_rank);

  end_work_unit = mpi_ct.local_work_units[mpi_ct.my_rank] - 1;
  for (iw = 0; iw <= end_work_unit; iw++) {
    MPI::Status status;

    mpi_ct.comm.Recv(&rs_local->view[iw]->ndet, 1, MPI::INT, 0, 0, status);
    mpi_ct.comm.Recv(&rs_local->view[iw]->view_angle, 1, MPI::DOUBLE, 0, 0, status);
    mpi_ct.comm.Recv(rs_local->view[iw]->detval, rs_local->ndet, MPI::FLOAT, 0, 0, status);
  }
  rs_local->nview = mpi_ct.local_work_units[mpi_ct.my_rank];
  if (opt_debug) {
    mpi_ct.comm.Barrier();
    fprintf(stdout, "Done with mpi_scatter_rs in process %2d\n", mpi_ct.my_rank);
  }
}

#endif

static void print_raysum_info(const RAYSUM *rs)
{
  printf ("Number of detectors: %d\n", rs->ndet);
  printf ("    Number of views: %d\n", rs->nview);
  printf ("             Remark: %s\n", rs->remark);
  printf ("             phmlen: %f\n", rs->phmlen);
  printf ("          det_start: %f\n", rs->det_start);
  printf ("            det_inc: %f\n", rs->det_inc);
  printf ("          rot_start: %f\n", rs->rot_start);
  printf ("            rot_inc: %f\n", rs->rot_inc);
}

#ifndef NO_MAIN
int 
main (const int argc, char *const argv[])
{
  return (ctrec_main(argc, argv));
}
#endif

